/*
 *      lark-main-window.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "lark-main-window.h"
#include "lark-export-dialog.h"
#include "lark-eq-dialog.h"

enum {
	PROP_0,
	PROP_APP,
	PROP_GSTREAMER,
	PROP_CORE,
	PROP_DOCUMENT_LOADER,
	PROP_PREFERENCES_SETTINGS,
	PROP_EQUALIZER_SETTINGS,
	PROP_EXPORT_SETTINGS,
	PROP_FILENAMES,
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

struct _LarkMainWindow
{
	GtkApplicationWindow parent_instance;
	/*Objects*/
	GtkApplication *app;
	LarkGstreamer *gstreamer;
	LarkCore *core;
	LarkDocumentLoader *docloader;
	GSettings *prefsettings;
	GSettings *eqsettings;
	GSettings *expsettings;
	gchar **filenames;
	GspellTextView *spellview;
	/*Widgets*/
	GtkTextView *LarkTextView;
	GtkLabel *LarkHeaderBarCaption;
	GtkProgressBar *LarkHeaderBarProgress;
	GtkImage *LarkReadButtonImage;
	GtkImage *LarkExportButtonImage;
	GtkCheckMenuItem *LarkWindowMenuSpellchecker;
	/*Data*/
	gchar *caption;
	gint startline;
	gint startoffset;
	gint endline;
	gint endoffset;
	gint currentline;
	gboolean processing;
	gboolean spellcheckerused;
	gboolean reading;
	GError *lasterror;
	guint timerid;
	GPtrArray *lines;
	GPtrArray *marks;
	/*Signals*/
	gulong textreadyhandler;
	gulong textfeedhandler;
	gulong finishedhandler;
	gulong winstatehandler;
};

struct _LarkMainWindowClass
{
	GtkApplicationWindowClass parent_class;
};

G_DEFINE_TYPE (LarkMainWindow, lark_main_window, GTK_TYPE_APPLICATION_WINDOW)

static void lark_main_window_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkMainWindow *self = (LarkMainWindow *)object;

	switch (prop_id) {
		case PROP_APP:
			g_value_set_pointer(value, self->app);
			break;
		case PROP_GSTREAMER:
			g_value_set_pointer(value, self->gstreamer);
			break;
		case PROP_CORE:
			g_value_set_pointer(value, self->core);
			break;
		case PROP_DOCUMENT_LOADER:
			g_value_set_pointer(value, self->docloader);
			break;
		case PROP_PREFERENCES_SETTINGS:
			g_value_set_pointer(value, self->prefsettings);
			break;
		case PROP_EQUALIZER_SETTINGS:
			g_value_set_pointer(value, self->eqsettings);
			break;
		case PROP_EXPORT_SETTINGS:
			g_value_set_pointer(value, self->expsettings);
			break;
		case PROP_FILENAMES:
			g_value_set_pointer(value, self->filenames);
			break;
	}
}

static void lark_main_window_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkMainWindow *self = (LarkMainWindow *)object;

	switch (prop_id) {
		case PROP_APP:
			self->app = g_value_get_pointer(value);
			break;
		case PROP_GSTREAMER:
			self->gstreamer = g_value_get_pointer(value);
			break;
		case PROP_CORE:
			self->core = g_value_get_pointer(value);
			break;
		case PROP_DOCUMENT_LOADER:
			self->docloader = g_value_get_pointer(value);
			break;
		case PROP_PREFERENCES_SETTINGS:
			self->prefsettings = g_value_get_pointer(value);
			break;
		case PROP_EQUALIZER_SETTINGS:
			self->eqsettings = g_value_get_pointer(value);
			break;
		case PROP_EXPORT_SETTINGS:
			self->expsettings = g_value_get_pointer(value);
			break;
		case PROP_FILENAMES:
			self->filenames = g_value_get_pointer(value);
			break;
	}
}

static void lark_main_window_form_runtime_data(LarkMainWindow *self)
{
	GtkTextBuffer *buffer;
	GtkTextIter startiter, enditer;
	guint curline;
	gchar *line;
	GtkTextMark *mark;
	
	g_return_if_fail(self != NULL);
	
	if (self->LarkTextView != NULL) {
		buffer = gtk_text_view_get_buffer(self->LarkTextView);
		if (buffer != NULL) {
			/*Arrays with text lines and start marks*/
			self->lines = g_ptr_array_sized_new(self->endline - self->startline);
			self->marks = g_ptr_array_sized_new(self->endline - self->startline);
			/*Save lines and marks for future access*/
			for (curline = self->startline; curline <= self->endline; curline++) {
				/*Start of line*/
				if (curline != self->startline) {
					gtk_text_buffer_get_iter_at_line(buffer, &startiter, curline);
				} else {
					gtk_text_buffer_get_iter_at_line_offset(buffer, &startiter, curline, self->startoffset);
				}
				/*End of line*/
				if (curline != self->endline) {
					gtk_text_buffer_get_iter_at_line(buffer, &enditer, curline + 1);
				} else {
					gtk_text_buffer_get_iter_at_line_offset(buffer, &enditer, curline, self->endoffset);
				}
				/*Save line*/
				line = gtk_text_buffer_get_text(buffer, &startiter, &enditer, FALSE);
				g_ptr_array_add(self->lines, (gpointer)line);
				/*Save mark*/
				mark = gtk_text_buffer_create_mark(buffer, NULL, (const GtkTextIter *)&startiter, TRUE);
				g_ptr_array_add(self->marks, (gpointer)mark);
			}
		}
	}
}

static void lark_main_window_cleanup_runtime_data(LarkMainWindow *self)
{
	GtkTextBuffer *buffer;
	guint i;
	gchar *line;
	GtkTextMark *mark;
	
	g_return_if_fail(self != NULL);
	
	if (self->lines != NULL) {
		for (i = 0; i < self->lines->len; i++) {
			line = (gchar *)g_ptr_array_index(self->lines, i);
			g_free(line);
		}
		g_ptr_array_free(self->lines, TRUE);
		self->lines = NULL;
	}
	
	if (self->marks != NULL) {
		if (self->LarkTextView != NULL) {
			buffer = gtk_text_view_get_buffer(self->LarkTextView);
			if (buffer != NULL) {
				for (i = 0; i < self->marks->len; i++) {
					mark = (GtkTextMark *)g_ptr_array_index(self->marks, i);
					gtk_text_buffer_delete_mark(buffer, mark);
				}
			}
		}
		g_ptr_array_free(self->marks, TRUE);
		self->marks = NULL;
	}
}

static gboolean lark_main_window_update_progress(LarkMainWindow *self)
{
	gfloat progress;
	GtkTextBuffer *buffer;
	GtkTextMark *mark;
	
	g_return_val_if_fail(self != NULL, FALSE);
	
	/*Calculate percentage*/
	progress = 1.0 - ((self->endline - self->currentline) / (gfloat)(self->endline - self->startline));
	gtk_progress_bar_set_fraction(self->LarkHeaderBarProgress, progress);
	/*Update text position*/
	if ((self->marks != NULL) && (self->currentline - self->startline < self->marks->len)) {
		if (self->LarkTextView != NULL) {
			buffer = gtk_text_view_get_buffer(self->LarkTextView);
			if (buffer != NULL) {
				mark = (GtkTextMark *)g_ptr_array_index(self->marks, self->currentline - self->startline);
				if ((mark != NULL) && (!gtk_text_mark_get_deleted(mark))) {
					gtk_text_view_scroll_mark_onscreen(self->LarkTextView, mark);
				}
			}
		}
	}
	
	return FALSE;
}

static gboolean lark_main_window_update_detailed_progress(LarkMainWindow *self)
{
	GString *progstr;
	guint64 t;
	
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(self->caption != NULL, FALSE);
	
	/*Start with cached caption*/
	progstr = g_string_new((const gchar *)self->caption);
	/*Add percentage*/
	g_string_append_printf(progstr, " (%.0f %%, ", (1.0 - ((self->endline - self->currentline) / (gfloat)(self->endline - self->startline))) * 100.0);
	/*Full size*/
	g_string_append_printf(progstr, "%.0f Mb, ", lark_gstreamer_get_full_size(self->gstreamer) / (1000.0 * 1000.0));
	/*And full time*/
	t = lark_gstreamer_get_full_duration(self->gstreamer) / 3600;
	if (t > 0) {
		g_string_append_printf(progstr, "%" G_GUINT64_FORMAT ":", t);
	}
	t = (lark_gstreamer_get_full_duration(self->gstreamer) % 3600) / 60;
	if (t < 10) {
		g_string_append_printf(progstr, "0%" G_GUINT64_FORMAT ":", t);
	} else {
		g_string_append_printf(progstr, "%" G_GUINT64_FORMAT ":", t);
	}
	t = lark_gstreamer_get_full_duration(self->gstreamer) % 60;
	if (t < 10) {
		g_string_append_printf(progstr, "0%" G_GUINT64_FORMAT ")", t);
	} else {
		g_string_append_printf(progstr, "%" G_GUINT64_FORMAT ")", t);
	}
	/*Update caption*/
	gtk_label_set_text(self->LarkHeaderBarCaption, progstr->str);
	/*Free resources*/
	g_string_free(progstr, TRUE);
	
	return TRUE;
}

static gboolean lark_main_window_finish(LarkMainWindow *self)
{
	GAction *action;
	GNotification *notification;
	gchar *message;

	g_return_val_if_fail(self != NULL, FALSE);

	/*Set flag*/
	self->processing = FALSE;
	/*Reset actions states and icons*/
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	gtk_image_set_from_icon_name(self->LarkReadButtonImage, "media-playback-start-symbolic", GTK_ICON_SIZE_BUTTON);
	gtk_image_set_from_icon_name(self->LarkExportButtonImage, "media-record-symbolic", GTK_ICON_SIZE_BUTTON);
	/*Hide progress bar*/
	gtk_widget_hide(GTK_WIDGET(self->LarkHeaderBarProgress));
	/*Cleanup runtime data*/
	lark_main_window_cleanup_runtime_data(self);
	/*Remove detailed info*/
	if (!self->reading) {
		/*Remove update timer*/
		g_source_remove(self->timerid);
		/*Reset caption*/
		gtk_label_set_text(self->LarkHeaderBarCaption, self->caption);
		/*Show notification*/
		notification = g_notification_new(_("Lark has finished it\'s work!"));
		message = g_strdup_printf(_("Audiobook is saved as \'%s\'. Feel free to export another one!"), lark_gstreamer_get_file_name(self->gstreamer));
		g_notification_set_body(notification, message);
		g_free(message);
		g_application_send_notification(G_APPLICATION(self->app), "finished", notification);
		g_object_unref(notification);
	}
	/*Set editable*/
	gtk_text_view_set_editable(self->LarkTextView, TRUE);
	
	return FALSE;
}

static gboolean lark_main_window_finish_with_error(LarkMainWindow *self)
{
	GAction *action;
	gchar *errormsg;
	GtkWidget *dialog;

	g_return_val_if_fail(self != NULL, FALSE);

	/*Set flag*/
	self->processing = FALSE;
	/*Reset actions states and icons*/
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), TRUE);
	gtk_image_set_from_icon_name(self->LarkReadButtonImage, "media-playback-start-symbolic", GTK_ICON_SIZE_BUTTON);
	gtk_image_set_from_icon_name(self->LarkExportButtonImage, "media-record-symbolic", GTK_ICON_SIZE_BUTTON);
	/*Hide progress bar*/
	gtk_widget_hide(GTK_WIDGET(self->LarkHeaderBarProgress));
	/*Cleanup runtime data*/
	lark_main_window_cleanup_runtime_data(self);
	/*Remove detailed info*/
	if (!self->reading) {
		/*Remove update timer*/
		g_source_remove(self->timerid);
		/*Reset caption*/
		gtk_label_set_text(self->LarkHeaderBarCaption, self->caption);
	}
	/*Show error message*/
	if (self->lasterror != NULL) {
		errormsg = g_strdup_printf(_("Error occured while document processing: %s"), self->lasterror->message);
		g_error_free(self->lasterror);
	} else {
		errormsg = g_strdup_printf(_("Error occured while document processing"));
	}
	dialog = gtk_message_dialog_new(GTK_WINDOW(self),
									GTK_DIALOG_MODAL | GTK_DIALOG_USE_HEADER_BAR,
									GTK_MESSAGE_ERROR,
									GTK_BUTTONS_CLOSE,
									"%s", errormsg);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	g_free(errormsg);
	/*Set editable*/
	gtk_text_view_set_editable(self->LarkTextView, TRUE);
	
	return FALSE;
}

static void lark_main_window_open_text(LarkMainWindow *self, gboolean alltext)
{
	GtkTextBuffer *buffer;
	GtkTextIter startiter, enditer;
	GtkTextMark *mark;
			
	g_return_if_fail(self != NULL);
	
	buffer = gtk_text_view_get_buffer(self->LarkTextView);
	if (buffer != NULL) {
		if ((gtk_text_buffer_get_selection_bounds(buffer, &startiter, &enditer)) && (!alltext)) {
			/*Only selected text*/
			self->startline = gtk_text_iter_get_line(&startiter);
			self->startoffset = gtk_text_iter_get_line_offset(&startiter);
			self->endline = gtk_text_iter_get_line(&enditer);
			self->endoffset = gtk_text_iter_get_line_offset(&enditer);
		} else {
			/*All text*/
			mark = gtk_text_buffer_get_insert(buffer);
			gtk_text_buffer_get_iter_at_mark(buffer, &startiter, mark);
			self->startline = gtk_text_iter_get_line(&startiter);
			self->startoffset = gtk_text_iter_get_line_offset(&startiter);
			gtk_text_buffer_get_end_iter(buffer, &enditer);
			self->endline = gtk_text_iter_get_line(&enditer);
			self->endoffset = gtk_text_iter_get_line_offset(&enditer);
		}
		
		lark_main_window_form_runtime_data(self);
		
		self->currentline = self->startline;
		
		g_idle_add((GSourceFunc)lark_main_window_update_progress, self);
	}
}

static void lark_main_window_text_ready(LarkDocumentLoader *self, gchar *chunk, guint len, LarkMainWindow *window)
{
	GtkTextBuffer *buffer;
	GtkTextIter iter;

	g_return_if_fail(window != NULL);

	buffer = gtk_text_view_get_buffer(window->LarkTextView);

	if (buffer != NULL) {
		gtk_text_buffer_get_end_iter(buffer, &iter);
		gtk_text_buffer_insert(buffer, &iter, chunk, -1);
	}
}

static gchar *lark_main_window_text_needed(LarkCore *self, LarkMainWindow *window)
{
	gchar *line;
	
	g_return_val_if_fail(window != NULL, NULL);
	g_return_val_if_fail(self != NULL, NULL);
	
	if (window->lines != NULL) {
		if (window->currentline - window->startline < window->lines->len) {
			/*Duplicating string as it will be freed automatically*/
			line = g_strdup((gchar *)g_ptr_array_index(window->lines, window->currentline - window->startline));
			g_idle_add((GSourceFunc)lark_main_window_update_progress, window);
			window->currentline++;
			return line;
		}
	}
	
	return NULL;
}

static void lark_main_window_finished(LarkCore *self, gboolean successfully, GError *error, LarkMainWindow *window)
{
	g_return_if_fail(window != NULL);

	if (successfully) {
		g_idle_add((GSourceFunc)lark_main_window_finish, window);
	} else {
		window->lasterror = error;
		g_idle_add((GSourceFunc)lark_main_window_finish_with_error, window);
	}
}

static void lark_main_window_open_file(LarkMainWindow *self, gchar *filename)
{
	GAction *localaction;
	GtkTextBuffer *buffer;
	GtkTextIter startiter, enditer;
	gchar *errormsg;
	GError *error = NULL;
	GtkWidget *dialog;
	gchar *wincaption;

	g_return_if_fail(self != NULL);
	g_return_if_fail(filename != NULL);

	/*Open file and load text*/
	buffer = gtk_text_view_get_buffer(self->LarkTextView);
	if (buffer != NULL) {
		localaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(localaction), FALSE);
		localaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(localaction), FALSE);
		/*Clear previous text*/
		gtk_text_buffer_get_bounds(buffer, &startiter, &enditer);
		gtk_text_buffer_delete(buffer, &startiter, &enditer);
		/*Enable controls*/
		if (lark_document_loader_open(self->docloader, filename, &error)) {
			localaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
			g_simple_action_set_enabled(G_SIMPLE_ACTION(localaction), TRUE);
			localaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
			g_simple_action_set_enabled(G_SIMPLE_ACTION(localaction), TRUE);
			/*Place cursor at the beginning*/
			gtk_text_buffer_get_start_iter(buffer, &startiter);
			gtk_text_buffer_place_cursor(buffer, &startiter);
			/*Update window caption*/
			if (self->caption != NULL) {
				g_free(self->caption);
			}
			wincaption = lark_document_loader_get_title(self->docloader);
			if (wincaption != NULL) {
				self->caption = g_strdup(wincaption);
			} else {
				self->caption = lark_document_loader_get_filename(self->docloader);
			}
			gtk_label_set_text(self->LarkHeaderBarCaption, self->caption);
		} else {
			if (error != NULL) {
				errormsg = g_strdup_printf(_("Unable to load file \'%s\': %s"), filename, error->message);
				g_error_free(error);
			} else {
				errormsg = g_strdup_printf(_("Unable to load file \'%s\'"), filename);
			}
			/*Show error message*/
			dialog = gtk_message_dialog_new(GTK_WINDOW(self),
											GTK_DIALOG_MODAL | GTK_DIALOG_USE_HEADER_BAR,
											GTK_MESSAGE_ERROR,
											GTK_BUTTONS_CLOSE,
											"%s", errormsg);

			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
			g_free(errormsg);
			/*Update window caption*/
			if (self->caption != NULL) {
				g_free(self->caption);
			}
			self->caption = NULL;
			gtk_label_set_text(self->LarkHeaderBarCaption, _("Lark"));
		}
	}
}

static void lark_main_window_open_activated(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	LarkMainWindow *self = (LarkMainWindow *)data;
	GtkWidget *dialog;
	GtkFileFilter *filter;
	gint result;
	gchar *filename = NULL;

	dialog = gtk_file_chooser_dialog_new(_("Open File..."), GTK_WINDOW(self),
										GTK_FILE_CHOOSER_ACTION_OPEN,
										_("_Cancel"), GTK_RESPONSE_CANCEL,
										_("_Open"), GTK_RESPONSE_ACCEPT,
										NULL);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("All supported documents"));
	gtk_file_filter_add_pattern(filter, "*.txt");
	gtk_file_filter_add_pattern(filter, "*.pdf");
	gtk_file_filter_add_pattern(filter, "*.epub");
	gtk_file_filter_add_pattern(filter, "*.fb2");
	gtk_file_filter_add_pattern(filter, "*.fb2.zip");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("Text document"));
	gtk_file_filter_add_pattern(filter, "*.txt");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("PDF document"));
	gtk_file_filter_add_pattern(filter, "*.pdf");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	
	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("EPUB document"));
	gtk_file_filter_add_pattern(filter, "*.epub");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("FB2 document"));
	gtk_file_filter_add_pattern(filter, "*.fb2");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	
	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, _("Compressed FB2 document"));
	gtk_file_filter_add_pattern(filter, "*.fb2.zip");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
	
	result = gtk_dialog_run(GTK_DIALOG(dialog));

	if (result == GTK_RESPONSE_ACCEPT) {
		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	}

	gtk_widget_destroy (dialog);

	/*Open file and load text*/
	if (filename != NULL) {
		lark_main_window_open_file(self, filename);
		g_free(filename);
	}
}

/*static void save_activated(GSimpleAction *action, GVariant *parameter, LarkMainWindow *self)
{

}*/

static void lark_main_window_read_activated(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	LarkMainWindow *self = (LarkMainWindow *)data;
	GAction *curaction;

	if (!self->processing) {
		/*Prepare selected text*/
		lark_main_window_open_text(self, FALSE);
		/*Show progress bar*/
		gtk_widget_show(GTK_WIDGET(self->LarkHeaderBarProgress));
		/*Set editable*/
		gtk_text_view_set_editable(self->LarkTextView, FALSE);
		/*Start processing*/
		lark_core_start(self->core, FALSE);
		/*Update controls*/
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), FALSE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), FALSE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), FALSE);
		gtk_image_set_from_icon_name(self->LarkReadButtonImage, "media-playback-stop-symbolic", GTK_ICON_SIZE_BUTTON);
		/*Set flags*/
		self->processing = TRUE;
		self->reading = TRUE;
	} else {
		/*Stop processing*/
		lark_core_stop(self->core);
		/*Update controls*/
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		gtk_image_set_from_icon_name(self->LarkReadButtonImage, "media-playback-start-symbolic", GTK_ICON_SIZE_BUTTON);
		/*Cleanup runtime data*/
		lark_main_window_cleanup_runtime_data(self);
		/*Set editable*/
		gtk_text_view_set_editable(self->LarkTextView, TRUE);
		/*Hide progress bar*/
		gtk_widget_hide(GTK_WIDGET(self->LarkHeaderBarProgress));
		/*Set flags*/
		self->processing = FALSE;
		self->reading = FALSE;
	}
}

static void lark_main_window_export_started(GtkWidget *exportdialog, LarkMainWindow *self)
{
	GAction *action;

	/*Prepare selected text*/
	lark_main_window_open_text(self, TRUE);
	/*Show progress bar*/
	gtk_widget_show(GTK_WIDGET(self->LarkHeaderBarProgress));
	/*Set editable*/
	gtk_text_view_set_editable(self->LarkTextView, FALSE);
	/*Start processing*/
	lark_core_start(self->core, TRUE);
	/*Update controls*/
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), FALSE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), FALSE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), FALSE);
	gtk_image_set_from_icon_name(self->LarkExportButtonImage, "media-playback-stop-symbolic", GTK_ICON_SIZE_BUTTON);
	/*Set flags*/
	self->processing = TRUE;
	self->reading = FALSE;
	/*Start timer to update detailed progress information*/
	self->timerid = g_timeout_add_seconds(1, (GSourceFunc)lark_main_window_update_detailed_progress, self);
}

static void lark_main_window_export_activated(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	LarkMainWindow *self = (LarkMainWindow *)data;
	GAction *curaction;
	GtkWidget *exportdialog;

	if (!self->processing) {
		/*Open dialog to set export parameters*/
		exportdialog = lark_export_dialog_new(GTK_WINDOW(self), self->gstreamer, self->docloader, self->expsettings);
		g_signal_connect(G_OBJECT(exportdialog), "export-started", (GCallback)lark_main_window_export_started, self);
		gtk_window_present(GTK_WINDOW(exportdialog));
	} else {
		/*Stop processing*/
		lark_core_stop(self->core);
		/*Update controls*/
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "open");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		curaction = g_action_map_lookup_action(G_ACTION_MAP(self->app), "spellchecker");
		g_simple_action_set_enabled(G_SIMPLE_ACTION(curaction), TRUE);
		gtk_image_set_from_icon_name(self->LarkExportButtonImage, "media-record-symbolic", GTK_ICON_SIZE_BUTTON);
		/*Cleanup runtime data*/
		lark_main_window_cleanup_runtime_data(self);
		/*Set editable*/
		gtk_text_view_set_editable(self->LarkTextView, TRUE);
		/*Hide progress bar*/
		gtk_widget_hide(GTK_WIDGET(self->LarkHeaderBarProgress));
		/*Set flags*/
		self->processing = FALSE;
		self->reading = FALSE;
		/*Remove update timer*/
		g_source_remove(self->timerid);
		/*Reset caption*/
		gtk_label_set_text(self->LarkHeaderBarCaption, self->caption);
	}
}

static void lark_main_window_equalizer_activated(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	LarkMainWindow *self = (LarkMainWindow *)data;
	GtkWidget *eqdialog;

	eqdialog = lark_eq_dialog_new(GTK_WINDOW(self), self->gstreamer, self->eqsettings);
	gtk_window_present(GTK_WINDOW(eqdialog));
}

static void lark_main_window_spellchecker_activated(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	LarkMainWindow *self = (LarkMainWindow *)data;

	if (self->spellcheckerused) {
		gtk_check_menu_item_set_active(self->LarkWindowMenuSpellchecker, FALSE);
		gspell_text_view_set_inline_spell_checking(self->spellview, FALSE);
		gspell_text_view_set_enable_language_menu(self->spellview, FALSE);
		self->spellcheckerused = FALSE;
	} else {
		gtk_check_menu_item_set_active(self->LarkWindowMenuSpellchecker, TRUE);
		gspell_text_view_set_inline_spell_checking(self->spellview, TRUE);
		gspell_text_view_set_enable_language_menu(self->spellview, TRUE);
		self->spellcheckerused = TRUE;
	}
}

static GActionEntry app_entries[] =
{
	{ "open",         lark_main_window_open_activated,         NULL, NULL, NULL },
	{ "read",         lark_main_window_read_activated,         NULL, NULL, NULL },
	{ "export",       lark_main_window_export_activated,       NULL, NULL, NULL },
	{ "spellchecker", lark_main_window_spellchecker_activated, NULL, NULL, NULL },
	{ "equalizer",    lark_main_window_equalizer_activated,    NULL, NULL, NULL }
};

static void lark_main_window_constructed(GObject *object)
{
	LarkMainWindow *self = (LarkMainWindow *)object;
	GAction *action;
	GspellChecker *checker;
	GtkTextBuffer *buffer;
	GspellTextBuffer *spellbuffer;

	/*Declare actions*/
	g_action_map_add_action_entries(G_ACTION_MAP(self->app), app_entries, G_N_ELEMENTS(app_entries), self);
	/*Disable controls*/
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "read");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), FALSE);
	action = g_action_map_lookup_action(G_ACTION_MAP(self->app), "export");
	g_simple_action_set_enabled(G_SIMPLE_ACTION(action), FALSE);

	/*Spellchecker*/
	checker = gspell_checker_new(NULL);
	buffer = gtk_text_view_get_buffer(self->LarkTextView);
	spellbuffer = gspell_text_buffer_get_from_gtk_text_buffer(buffer);
	gspell_text_buffer_set_spell_checker(spellbuffer, checker);
	g_object_unref(checker);
	
	self->spellview = gspell_text_view_get_from_gtk_text_view(self->LarkTextView);
	gspell_text_view_set_inline_spell_checking(self->spellview, FALSE);
	gspell_text_view_set_enable_language_menu(self->spellview, FALSE);
	
	/*Bind signals*/
	self->textreadyhandler = g_signal_connect(self->docloader, "text-chunk-ready", (GCallback)lark_main_window_text_ready, self);
	self->textfeedhandler = g_signal_connect(self->core, "text-needed", (GCallback)lark_main_window_text_needed, self);
	self->finishedhandler = g_signal_connect(self->core, "finished", (GCallback)lark_main_window_finished, self);

	/*Open file and load text*/
	if (self->filenames != NULL) {
		if (self->filenames[0] != NULL) {
			lark_main_window_open_file(self, self->filenames[0]);
		}
	}

	G_OBJECT_CLASS(lark_main_window_parent_class)->constructed(object);
}

static void lark_main_window_finalize(GObject *object)
{
	LarkMainWindow *self = (LarkMainWindow *)object;

	/*Stop playback*/
	lark_core_stop(self->core);

	/*Unbind signals*/
	if (self->finishedhandler != 0) {
		g_signal_handler_disconnect(self->core, self->finishedhandler);
	}
	if (self->textfeedhandler != 0) {
		g_signal_handler_disconnect(self->core, self->textfeedhandler);
	}
	if (self->textreadyhandler != 0) {
		g_signal_handler_disconnect(self->docloader, self->textreadyhandler);
	}
	
	G_OBJECT_CLASS(lark_main_window_parent_class)->finalize(object);
}

static void lark_main_window_class_init(LarkMainWindowClass *klass)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
	GObjectClass *object_class = G_OBJECT_CLASS(klass);

	object_class->get_property = lark_main_window_get_property;
	object_class->set_property = lark_main_window_set_property;
	object_class->constructed = lark_main_window_constructed;
	object_class->finalize = lark_main_window_finalize;

	properties[PROP_APP] =
		g_param_spec_pointer("app",
							"Application object",
							"Pointer to application object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_GSTREAMER] =
		g_param_spec_pointer("gstreamer",
							"GStreamer object",
							"Pointer to GStreamer object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_CORE] =
		g_param_spec_pointer("core",
							"Core object",
							"Pointer to core object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_DOCUMENT_LOADER] =
		g_param_spec_pointer("docloader",
							"Document loader object",
							"Pointer to document loader object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_PREFERENCES_SETTINGS] =
		g_param_spec_pointer("preferences-settings",
							"Preferences settings object",
							"Pointer to preferences settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_EQUALIZER_SETTINGS] =
		g_param_spec_pointer("equalizer-settings",
							"Equalizer settings object",
							"Pointer to equalizer settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_EXPORT_SETTINGS] =
		g_param_spec_pointer("export-settings",
							"Export settings object",
							"Pointer to export settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	properties[PROP_FILENAMES] =
		g_param_spec_pointer("filenames",
							"Pointer to filenames array",
							"Pointer to filenames array",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_properties(object_class, LAST_PROP, properties);

	gtk_widget_class_set_template_from_resource(widget_class, "/ru/linuxonly/lark/ui/main-window.ui");

	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkTextView);
	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkHeaderBarCaption);
	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkHeaderBarProgress);
	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkReadButtonImage);
	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkExportButtonImage);
	gtk_widget_class_bind_template_child(widget_class, LarkMainWindow, LarkWindowMenuSpellchecker);

	//gtk_widget_class_bind_template_callback(widget_class, lark_main_window_screen_button_press_event);
}

static void lark_main_window_init(LarkMainWindow *self)
{
	gtk_widget_init_template(GTK_WIDGET(self));
	
	self->caption = NULL;
	self->processing = FALSE;
	self->reading = FALSE;
	self->spellcheckerused = FALSE;
	self->startline = 0;
	self->startoffset = 0;
	self->endline = 0;
	self->endoffset = 0;
	self->currentline = 0;
	self->lines = NULL;
	self->marks = NULL;
}

GtkWidget *lark_main_window_new(LarkApp *app, LarkGstreamer *gstreamer, LarkCore *core, LarkDocumentLoader *docloader, GSettings *prefsettings, GSettings *eqsettings, GSettings *expsettings, gchar **filenames)
{
	return g_object_new(lark_main_window_get_type(), "application", app, "app", app, "gstreamer", gstreamer, "core", core, "docloader", docloader, "preferences-settings", prefsettings, "equalizer-settings", eqsettings, "export-settings", expsettings, "filenames", filenames, NULL);
}
